/* content of dxf2html.css */
body {
    font-family: "Open Sans", sans-serif;
    width: 1100px;
    margin-right: auto;
    margin-left: auto;
    color: white;
    background-color: lightgray;
}

#dxf-filename {
    color: #20324D;
    font-size: xx-large;
    font-weight: bold;
    padding: 5px;
}

/* Background */
.dxf-bg {
    background-color: #20324D;
    padding: 5px;
}

.dxf-tag-group {
    background-color: #3F5577;
    margin: 5px 0px 5px 0px;
    padding: 5px 5px 5px 5px;
}

.tag-group-marker {
    color: #A0BFDE;
    /* font-weight: bold; */
}

.tag-control-tag {
    color: orange;
    font-weight: bold;
}

.dxf-tag {
    font-family: "Source Code Pro", monospace;
    font-size: medium;
}

.tag-code {
}

.var-type {
}

.tag-value {
}

